import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;


@SuppressWarnings("unchecked")
public class InvoiceEntry implements Comparable
{
	// create the Date Format, EX: "Wed Feb 9, 2011 1:30am UTC" OR "Wed Feb 9, 2011 1am UTC"
	// formats[2] is used when PRINTING the date objects
	private static final SimpleDateFormat[] formats = {new SimpleDateFormat("EEE MMM d, yyyy hh:mmaaa z"), 
													   new SimpleDateFormat("EEE MMM d, yyyy hhaaa z"),
													   new SimpleDateFormat("EEE MMM d, yyyy hh:mmaaa")};
	
	private String title;
	private String description;
	private String location;
	private float hours;
	private String date;
	private Date startDate;
	private Date endDate;
	private CalendarReader parent;
	
	public InvoiceEntry()
	{
		title = null;
		description = null;
		location = null;
		hours = 0;
		date = null;
		startDate = null;
		endDate = null;
	}
	
	public InvoiceEntry(String title, String description, String location, String date, CalendarReader p) throws DateFormatError
	{
		setTitle(title);
		setDescription(description);
		setLocation(location);
		setDateText(date);
		parent = p;
	}
	
	// -------------------------------------------------------------------------
	// CLASS FUNCTIONALITY *****************************************************
	
	// basic textfile layout
	public String print()
	{
		String retVal = getTitle() + '\n';
		retVal += "-------------" + '\n';
		retVal += getFormatedDateText();
		retVal += "   @" + getLocation() + '\n' + '\n';
		retVal += getDescription() + '\n' + '\n';
		retVal += "HOURS WORKED: " + getHours() + '\n';
		
		return retVal;
	}
	
	// this print method takes an InvoiceEntryFormat object and prints only that information
	// specified by the filter
	public String print(InvoiceEntryFormat ie)
	{
		String retVal = new String();
		
		if(ie.getTitle())
		{
			retVal += getTitle() + '\n';
		}
		if(ie.getFormatedDateText())
		{
			retVal += getFormatedDateText() + '\n';
		}
		if(ie.getLocation())
		{
			retVal += '@' + getLocation() + '\n';
		}
		if(ie.getDescription())
		{
			retVal += getDescription() + '\n';
		}
		if(ie.getHours())
		{
			retVal += "Total Hours: " + getHours() + '\n';
		}
		
		return retVal;
	}
	
	public String printHTML(InvoiceEntryFormat ie)
	{
		String retVal = "<div class=\"entry\">";
		
		if(ie.getTitle())
		{
			retVal += "<div class=\"entry_title bold\"><p>" + getTitle() + "</p></div>";
		}
		if(ie.getFormatedDateText())
		{
			retVal += "<div class=\"entry_date italic\"><p>" + getFormatedDateText() + "</p></div>";
		}
		if(ie.getLocation())
		{
			retVal += "<div class=\"entry_location\"><p>" + '@' + getLocation() + "</p></div>";
		}
		if(ie.getDescription())
		{
			retVal += "<div class=\"entry_description\"><p>" + getDescription() + "</p></div>";
		}
		if(ie.getHours())
		{
			retVal += "<div class=\"entry_hours bold\"><p>Total Hours:" + getHours() + "</p></div>";
		}
		
		retVal += "</div>";
		
		return retVal;
	}
	
	// stylized sheet in HTML format
	public String printToHTML()
	{
		String retVal = null;
		
		retVal = "<tr><td width='535px' >"		
						+ "<h2>" + getTitle() + "</h2>"
						+ "<p>" + getDateText() + "</p>" + getDescription() + "<br><br><strong>hours: </strong>" + getHours() 
						+ "<br><br><br></td></tr><tr><td height='20px'></td></tr>";
		
		return retVal;
	}
	
	// -------------------------------------------------------------------------
	// GETTER FUNCTIONS ********************************************************
	public String getTitle()
	{
		return title;
	}
	
	public String getDescription()
	{
		return description;
	}
	
	public String getLocation()
	{
		return location;
	}
	
	public String getDateText()
	{
		return date;
	}
	
	public float getHours()
	{
		return hours;
	}
	
	public Date getStartDate()
	{
		return startDate;
	}
	
	public Date getEndDate()
	{
		return endDate;
	}
	
	public String getFormatedDateText()
	{
		String retVal = new String();
		
		retVal = formats[2].format(getStartDate()) + " to " + formats[2].format(getEndDate());
		
		return retVal;
	}
	
	// -------------------------------------------------------------------------
	// SETTER FUNCTIONS ********************************************************
	
	public void setTitle(String title)
	{
		this.title = title;
	}
	
	public void setDescription(String desc)
	{
		this.description = desc;
	}
	
	public void setLocation(String location)
	{
		this.location = location;
	}
	
	public void setDateText(String date) throws DateFormatError
	{
		this.date = date;
		extractDates();
	}
	
	private void setHours()
	{
		this.hours = ((float)(getEndDate().getTime() - getStartDate().getTime())/(float)(1000*60*60));
	}
	
	private void setStartDate(Date date)
	{
		this.startDate = date;
	}
	
	private void setEndDate(Date date)
	{
		this.endDate = date;
	}
	
	private void extractDates() throws DateFormatError
	{
		// EXAMPLE DATETEXT FORMATS:
		// Thu Oct 7, 2010 12am to 2amUTC
		// Sun Oct 31, 2010 11pm to Mon Nov 1, 2010 12amUTC
		
		// Split the two dates slong on the " to "  string
		String [] dates = getDateText().split(" to ");
		
		// extract the timezone (always 3 characters from the end of the second date)
		String zone = dates[1].substring(dates[1].length()-3);
		
		// concatenate the second date to entirely remove the timezone stamp and the space before it - " UTC"
		dates[1] = dates[1].substring(0, dates[1].length()-4);
		
		// create temporary date holder objects
		Date date0 = null;
		Date date1 = null;
		
		// attempt to parse the first date (dates[0])
		date0 = parseDateText(dates[0]+ " " + zone);
		
		try 
		{
			// attempt to parse the second date (dates[1])
			date1 = parseDateText(dates[1] + " " + zone);
			//System.out.println("DOUBLE DATED, " + getDateText());
		} 
		// if we catch a DateFormatError we will attempt to refactor the date and try agian, see below
		catch (DateFormatError e) 
		{
			// System.out.println("SINGLE DATED, " + getDateText());
			// if there is an error, we assume it was because the two dates occurred on the same day
			// and therefore the dateText follows example 1: (Thu Oct 7, 2010 12am to 2amUTC)
			// with the second date string containing only "2am" instead of the full date information
			// I call this being "single dated" as opposed to being "double dated"
			
			// to compensate for this, we grab the first date's long string information (except the time) 
			String date1Information = dates[0].substring(0,dates[0].lastIndexOf(" "));
			
			// we then append that information to the beginning of the second dates
			dates[1] = date1Information + " " + dates[1];
			
			// attempt to parse the second date again (dates[1])
			date1 = parseDateText(dates[1]+ " " + zone);
		}
		
		/* TODO: 
			NOTE: "double dating" is my term for when google writes this:
			Sun Oct 31, 2010 11pm to Mon Nov 1, 2010 12amUTC
			
			as opposed to "sinlge dating":
			Thu Oct 7, 2010 12am to 2amUTC
 
			For some reason, Google basis it's double dating standard *REQUIRES MORE TESTING TO CONFIM* 
			based on UTC time. So, if based on UTC time the two timestamps occur on the same day then
			then the date string is single dated. This causes a problem because in cases where the
			timezone is different (IE: CDT) then the string (which should be double dated by CDT standards)
			wont be and will read Day 1 8:30pm - 12:30am instead of	Day 1 8:30pm - Day 2 12:30am. This 
			throws a kink into the extraction algorithm which assumes when the date is single dated then 
			the two days match. The result is a second date that precedes the first date. 
			So, to combat this, we must check and make the appropriate adjustment
		*/ 
		
		// if the second date comes before the first date
		if(date0.after(date1))
		{
			// we add one full day to the timestamp
			long newTime = date1.getTime()+(1000*60*60*24);
			
			// and re-create the second date to match the correct time
			date1 = new Date(newTime);
		}
		
		// at this point both temporary date holders are created, store them permanently
		setStartDate(date0);
		setEndDate(date1);
		
		// we now calculate the hours
		setHours();
	}
	
	// this function takes a string and determines if it fits the date formatting styles
	// if it does it returns a working Date object, if it does not an exception is thrown
	private Date parseDateText(String text) throws DateFormatError
	{
		Date retVal = null;
	
		// this loop iterates through the different date formats
		for(int i=0; i<2; i++)
		{
			try
			{
				// we attempt to parse the text,
				retVal = InvoiceEntry.formats[i].parse(text);
				
				// if it succeeds we break out of the loop
				break;
			}
			// if it fails and a ParseException is caught
			catch(ParseException e) 
			{
				// we ensure retVal equals null so that the next file formatting can be attempted
				retVal = null;
			}
		}
		
		// if retVal is null then the string was designated as not a valid date text
		if(retVal == null)
		{
			// we throw a DateFormatError and pass in the improperly formatted text
			throw new DateFormatError(text);
		}
		
		// if we succeeded we return a working Date object
		return retVal;
	}

	// an invoice entry is compared to another invoice entry entirely based on its start date
	// whichever invoice entry has an earlier start date is described as being "less than" the
	// the object whose start date is later
	@Override
	public int compareTo(Object obj2) 
	{
		// cast the object as an InvoiceEntry
		InvoiceEntry other = (InvoiceEntry) obj2;
		
		// extract the start date of the two objects
		Date thisDate = this.getStartDate();
		Date otherDate = other.getStartDate();
		
		// the return value will be the result of the comparison of their start dates
		int retVal = thisDate.compareTo(otherDate);
		
		// return the result
		return retVal;
	}
	
	@Override
	public String toString()
	{
		return this.print();
	}
	
	// returns the CalendarReader from where this Entry came
	public CalendarReader getParent()
	{
		return parent;
	}
	
}
